# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Copyright 2023 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import typing as tp

import jax
import jax.tree_util as jtu

from flax import traverse_util
from flax.experimental.nnx.nnx import filterlib, reprlib
from flax.experimental.nnx.nnx.variables import Variable

A = tp.TypeVar('A')

Leaf = tp.Any
Path = str
Key = str
FlatState = dict[Path, Variable[Leaf]]


class StateVariablesMapping(
  tp.MutableMapping[str, Variable[tp.Any]], reprlib.Representable
):
  __slots__ = ('_mapping',)

  def __init__(self, mapping: dict[Key, tp.Any]):
    if tp.TYPE_CHECKING:
      self._mapping = mapping
    else:
      object.__setattr__(self, '_mapping', mapping)

  def __getitem__(self, key: str | int) -> Variable[tp.Any]:
    if isinstance(key, int):
      key = str(key)

    value = self._mapping[key]

    if not isinstance(value, Variable):
      raise KeyError(f"Variable '{key}' not found.")

    return value

  def __setitem__(self, name: str, value: Variable[tp.Any]) -> None:
    self._mapping[name] = value

  def __getattr__(self, name: str) -> Variable[tp.Any]:
    value = self._mapping[name]
    if not isinstance(value, Variable):
      raise AttributeError(f"Variable '{name}' not found.")
    return value

  def __setattr__(self, name: str, value: Variable[tp.Any]) -> None:
    self._mapping[name] = value

  def __delitem__(self, name: str) -> None:
    del self._mapping[name]

  def __iter__(self) -> tp.Iterator[str]:
    for name, value in self._mapping.items():
      if isinstance(value, Variable):
        yield name

  def __len__(self) -> int:
    return sum(1 for _ in self)

  def __nnx_repr__(self):
    yield reprlib.Object(type(self), start='{', end='}', value_sep=': ')
    for name, value in vars(self._mapping).items():
      if isinstance(value, Variable):
        yield reprlib.Attr(repr(name), value)


class NestedStateRepr(reprlib.Representable):
  def __init__(self, state: State):
    self.state = state

  def __nnx_repr__(self):
    yield reprlib.Object('', value_sep=': ', start='{', end='}')

    for r in self.state.__nnx_repr__():
      if isinstance(r, reprlib.Object):
        continue
      yield r


class State(tp.MutableMapping[Key, tp.Any], reprlib.Representable):
  def __init__(
    self,
    mapping: tp.Union[
      tp.Mapping[Key, tp.Any],
      tp.Iterator[tp.Tuple[Key, tp.Any]],
    ],
    /,
  ):
    if tp.TYPE_CHECKING:
      self._mapping = dict(mapping)
    else:
      super().__setattr__('_mapping', dict(mapping))

  @property
  def raw_mapping(self) -> dict[Key, dict[str, tp.Any] | tp.Any]:
    return self._mapping

  @property
  def variables(self) -> StateVariablesMapping:
    return StateVariablesMapping(self._mapping)

  def __getitem__(self, key: Key | int) -> Leaf | State:
    if isinstance(key, int):
      key = str(key)
    value = self._mapping[key]
    if isinstance(value, Variable):
      return value.value
    return State(value)

  def __getattr__(self, key: Key) -> Leaf | State:
    if '_mapping' not in vars(self) or key not in self._mapping:
      raise AttributeError(f'No attribute {key} in State')

    return self[key]

  def __setitem__(self, key: Key | int, value: Leaf | State) -> None:
    if isinstance(key, int):
      key = str(key)
    if isinstance(value, State):
      self._mapping[key] = value._mapping
    else:
      if not isinstance(self._mapping[key], Variable):
        raise ValueError(
          f'Trying to set key {key} to a leaf value '
          f'but current value is not a Variable: {self._mapping[key]}.'
        )
      self._mapping[key].value = value

  __setattr__ = __setitem__

  def __delitem__(self, key: Key) -> None:
    del self._mapping[key]

  def __iter__(self) -> tp.Iterator[Key]:
    return iter(self._mapping)

  def __len__(self) -> int:
    return len(self._mapping)

  def __nnx_repr__(self):
    yield reprlib.Object(type(self), value_sep=': ', start='({', end='})')

    for k, v in self.items():
      if isinstance(v, State):
        v = NestedStateRepr(v)
      yield reprlib.Attr(repr(k), v)

  def flat_state(self) -> dict[Key, Variable[Leaf]]:
    return traverse_util.flatten_dict(self._mapping, sep='/')  # type: ignore

  @classmethod
  def from_flat_path(cls, flat_state: FlatState) -> State:
    nested_state = traverse_util.unflatten_dict(flat_state, sep='/')
    return cls(nested_state)

  @tp.overload
  def split(self, first: filterlib.Filter, /) -> 'State':
    ...

  @tp.overload
  def split(
    self,
    first: filterlib.Filter,
    second: filterlib.Filter,
    /,
    *filters: filterlib.Filter,
  ) -> tp.Tuple['State', ...]:
    ...

  def split(
    self, first: filterlib.Filter, /, *filters: filterlib.Filter
  ) -> tp.Union['State', tp.Tuple['State', ...]]:
    filters = (first, *filters)
    *states, rest = _split_state(self, *filters)

    if rest:
      raise ValueError(
        'Non-exhaustive filters, got a non-empty remainder: '
        f'{list(rest.keys())}.\nUse `...` to match all remaining elements.'
      )

    if len(states) == 1:
      states = states[0]
    else:
      states = tuple(states)
    return states

  @tp.overload
  def extract(
    self,
    first: filterlib.Filter,
    /,
  ) -> 'State':
    ...

  @tp.overload
  def extract(
    self,
    first: filterlib.Filter,
    second: filterlib.Filter,
    /,
    *filters: filterlib.Filter,
  ) -> tp.Tuple['State', ...]:
    ...

  def extract(
    self,
    first: filterlib.Filter,
    /,
    *filters: filterlib.Filter,
  ) -> tp.Union['State', tp.Tuple['State', ...]]:
    *states, _rest = _split_state(self, first, *filters)

    assert len(states) == len(filters) + 1

    if len(states) == 1:
      states = states[0]
    else:
      states = tuple(states)

    return states

  @staticmethod
  def merge(state: 'State', /, *states: 'State') -> 'State':
    states = (state, *states)

    if len(states) == 1:
      return states[0]

    new_state: FlatState = {}

    for state in states:
      new_state.update(state.flat_state())

    return State.from_flat_path(new_state)

  def __or__(self, other: 'State') -> 'State':
    if not other:
      return self
    return State.merge(self, other)

  def __sub__(self, other: 'State') -> 'State':
    if not other:
      return self

    _mapping = {k: v for k, v in self._mapping.items() if k not in other}
    return State(_mapping)


def _state_flatten_with_keys(x: State):
  items = sorted(x._mapping.items(), key=lambda item: item[0])
  children = tuple((jtu.DictKey(key), value) for key, value in items)
  return children, tuple(x._mapping.keys())


def _state_unflatten(
  static: tp.Tuple[Path, ...] | None,
  leaves: tp.Tuple[Leaf, ...] | tuple[dict[str, Leaf]],
):
  return State(zip(static, leaves)) if static else State(leaves[0])


def _state_flatten(x: State):
  return (x._mapping,), None


jax.tree_util.register_pytree_with_keys(
  State,
  _state_flatten_with_keys,
  _state_unflatten,
  flatten_func=_state_flatten,
)


def _split_state(
  state: State,
  *filters: filterlib.Filter,
) -> tp.Tuple[State, ...]:
  for i, filter_ in enumerate(filters):
    if filter_ is ... and i != len(filters) - 1:
      raise ValueError(
        'Ellipsis `...` can only be used as the last filter, '
        f'got it at index {i}.'
      )
  predicates = tuple(map(filterlib.to_predicate, filters))

  flat_state = state.flat_state()

  # we have n + 1 states, where n is the number of predicates
  # the last state is for values that don't match any predicate
  flat_states: tp.Tuple[FlatState, ...] = tuple(
    {} for _ in range(len(predicates) + 1)
  )

  for path, value in flat_state.items():
    for i, predicate in enumerate(predicates):
      if predicate(path, value):
        flat_states[i][path] = value
        break
    else:
      # if we didn't break, set leaf to last state
      flat_states[-1][path] = value

  return tuple(State.from_flat_path(flat_state) for flat_state in flat_states)



# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import jax
import pytest

from flax.experimental import nnx


class TestPartitioning:
  def test_partition(self):
    m = nnx.Dict(
      a=nnx.Sequence([nnx.Param(1), nnx.BatchStat(2)]),
      b=nnx.Param(2),
      c=100,
    )

    params, rest, graphdef = m.split(nnx.Param, ...)

    assert len(params) == 2
    assert len(rest) == 1

    # check params
    assert params['a']['0'] == m.a[0]
    assert params['b'] == m.b

    # check rest
    assert rest['a']['1'] == m.a[1]

    m2 = graphdef.merge(params, rest)

    assert m2.a[0] == m.a[0]
    assert m2.a[1] == m.a[1]
    assert m2.b == m.b
    assert m2.c == 100

  def test_complete_partitioning(self):
    m = nnx.Dict(
      a=nnx.Sequence([nnx.Param(1), nnx.Param(2), nnx.Variable(3)]),
      b=nnx.Dict(c=nnx.Param(1), d=nnx.BatchStat(2)),
    )

    # no error
    m.split(nnx.Param, nnx.BatchStat, nnx.Variable)

  def test_complete_partitioning_plus_ellipsis(self):
    m = nnx.Dict(
      a=nnx.Sequence([nnx.Param(1), nnx.Param(2), nnx.Variable(3)]),
      b=nnx.Dict(c=nnx.Param(1), d=nnx.BatchStat(2)),
    )

    # no error if additional ... is passed at the end
    m.split(nnx.Param, nnx.BatchStat, nnx.Variable, ...)

  def test_inclomplete_partition_error(self):
    m = nnx.Dict(
      a=nnx.Sequence([nnx.Param(1), nnx.Param(2), nnx.Variable(3)]),
      b=nnx.Dict(c=nnx.Param(1), d=nnx.BatchStat(2)),
    )

    with pytest.raises(
      ValueError, match='Non-exhaustive filters, got a non-empty remainder'
    ):
      m.split(nnx.Param)

  def test_ellipsis_not_last_error(self):
    m = nnx.Dict(
      a=nnx.Sequence([nnx.Param(1), nnx.Param(2), nnx.Variable(3)]),
      b=nnx.Dict(c=nnx.Param(1), d=nnx.BatchStat(2)),
    )

    with pytest.raises(
      ValueError, match='Ellipsis `...` can only be used as the last filter,'
    ):
      m.split(..., nnx.Param)

  def test_update_from(self):
    m = nnx.Dict(
      a=nnx.Sequence([nnx.Param(1), nnx.BatchStat(3)]),
      b=nnx.Param(2),
      c=100,
    )

    state = m.split()[0]
    state = jax.tree_map(lambda x: x * 2, state)

    m.update(state)

    assert m.a[0] == 2
    assert m.a[1] == 6
    assert m.b == 4
    assert m.c == 100

  def test_update_from_with_array_leaf(self):
    m = nnx.Dict(
      a=nnx.Sequence([nnx.Param(1), nnx.BatchStat(3)]),
      b=nnx.Param(2),
      c=nnx.Variable(jax.numpy.array(100)),
    )

    state, graphdef = m.split()
    state = jax.tree_map(lambda x: x * 2, state)

    m.update(state)

    assert m.a[0] == 2
    assert m.a[1] == 6
    assert m.b == 4
    assert m.c == 200

  def test_grad_example(self):
    m = nnx.Dict(
      a=nnx.Sequence([nnx.Param(1.0), nnx.BatchStat(-10)]),
      b=nnx.Param(2.0),
      c=100,
    )

    params = m.extract(nnx.Param)

    def loss(params):
      return sum(2 * p for p in jax.tree_util.tree_leaves(params))

    grads = jax.grad(loss)(params)
    m.update(grads)

    assert m.a[0] == 2.0
    assert m.a[1] == -10
    assert m.b == 2.0
    assert m.c == 100

  def test_get_paritition(self):
    m = nnx.Dict(
      a=nnx.Sequence([nnx.Param(10.0), nnx.Param(20.0)]),
      b=nnx.Param(10.0),
      c=7,
      d=5.0,
    )

    # test Variables not shared
    assert vars(m.a)['0'] is not vars(m)['b']

    state = m.extract(nnx.Variable)
    assert state['a']['0'] == m.a[0]
    assert state['a']['1'] == m.a[1]
    assert state['b'] == m.b
    assert state.variables.b is not state.a.variables[0]
    assert len(state.flat_state()) == 3



# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Copyright 2021 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Default Hyperparameter configuration."""

import ml_collections


def get_config():
  """Get the default hyperparameter configuration."""
  config = ml_collections.ConfigDict()

  # As defined in the `models` module.
  config.model = 'ResNet50'
  # `name` argument of tensorflow_datasets.builder()
  config.dataset = 'imagenet2012:5.*.*'

  config.learning_rate = 0.1
  config.warmup_epochs = 5.0
  config.momentum = 0.9
  config.batch_size = 128
  config.shuffle_buffer_size = 16 * 128
  config.prefetch = 10

  config.num_epochs = 100.0
  config.log_every_steps = 100

  config.cache = False
  config.half_precision = False

  # If num_train_steps==-1 then the number of training steps is calculated from
  # num_epochs using the entire dataset. Similarly for steps_per_eval.
  config.num_train_steps = -1
  config.steps_per_eval = -1
  return config


def metrics():
  return [
      'train_loss',
      'eval_loss',
      'train_accuracy',
      'eval_accuracy',
      'steps_per_second',
      'train_learning_rate',
  ]



# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for flax.linen.dtypes."""


import jax
from absl.testing import absltest
from jax import numpy as jnp

from flax.linen import dtypes

default_float_dtype = jnp.result_type(1.0)


class DtypesTest(absltest.TestCase):
  def test_no_inexact_dtype(self):
    i32 = jnp.int32(1.0)
    self.assertEqual(dtypes.canonicalize_dtype(i32, inexact=False), jnp.int32)

  def test_inexact_dtype(self):
    with jax.experimental.enable_x64():
      i64 = jnp.int64(1)
      self.assertEqual(dtypes.canonicalize_dtype(i64), jnp.float32)
    i32 = jnp.int32(1)
    self.assertEqual(dtypes.canonicalize_dtype(i32), jnp.float32)
    i16 = jnp.int16(1.0)
    self.assertEqual(dtypes.canonicalize_dtype(i16), jnp.float32)

  def test_explicit_downcast(self):
    f32 = jnp.float32(1.0)
    (x,) = dtypes.promote_dtype(f32, dtype=jnp.float16)
    self.assertEqual(x.dtype, jnp.float16)


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional, Sequence

import jax.numpy as jnp
from jax import lax, random

from flax.experimental.nnx.nnx import dataclasses as nnx_dataclasses
from flax.experimental.nnx.nnx import flaglib, rnglib
from flax.experimental.nnx.nnx.module import Module, first_from


@nnx_dataclasses.dataclass
class Dropout(Module):
  """Create a dropout layer.

  Attributes:
    rate: the dropout probability.  (_not_ the keep rate!)
    broadcast_dims: dimensions that will share the same dropout mask
    deterministic: if false the inputs are scaled by `1 / (1 - rate)` and
      masked, whereas if true, no mask is applied and the inputs are returned
      as is.
    rng_collection: the rng collection name to use when requesting an rng key.
  """

  rate: float
  broadcast_dims: Sequence[int] = ()
  deterministic: Optional[bool] = None
  rng_collection: str = 'dropout'

  def __call__(
    self,
    inputs,
    *,
    deterministic: Optional[bool] = None,
    rngs: Optional[rnglib.Rngs] = None,
  ):
    """Applies a random dropout mask to the input.

    Args:
      inputs: the inputs that should be randomly masked.
      deterministic: if false the inputs are scaled by `1 / (1 - rate)` and
        masked, whereas if true, no mask is applied and the inputs are returned
        as is.

    Returns:
      The masked inputs reweighted to preserve mean.
    """
    deterministic = first_from(
      deterministic,
      self.deterministic,
      flaglib.flags.get('deterministic'),
      error_msg="""No `deterministic` argument was provided to Dropout
          as either a __call__ argument, class attribute, or nnx.flag.""",
    )

    if (self.rate == 0.0) or deterministic:
      return inputs

    # Prevent gradient NaNs in 1.0 edge-case.
    if self.rate == 1.0:
      return jnp.zeros_like(inputs)

    if rngs is None:
      raise ValueError(
        "Dropout needs to generate a random mask but no 'rngs' were provided."
      )

    keep_prob = 1.0 - self.rate
    rng = rngs[self.rng_collection]()
    broadcast_shape = list(inputs.shape)
    for dim in self.broadcast_dims:
      broadcast_shape[dim] = 1
    mask = random.bernoulli(rng, p=keep_prob, shape=broadcast_shape)
    mask = jnp.broadcast_to(mask, inputs.shape)
    return lax.select(mask, inputs / keep_prob, jnp.zeros_like(inputs))



# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import dataclasses
import typing as tp

import typing_extensions as tpe

from flax.experimental import nnx
from flax.experimental.nnx.nnx import pytreelib, variables

A = tp.TypeVar('A')


def field(
  *,
  default: tp.Any = dataclasses.MISSING,
  default_factory: tp.Any = dataclasses.MISSING,
  init: bool = True,
  repr: bool = True,
  hash: tp.Optional[bool] = None,
  compare: bool = True,
  metadata: tp.Optional[tp.Mapping[str, tp.Any]] = None,
):
  return dataclasses.field(  # type: ignore
    default=default,
    default_factory=default_factory,
    init=init,
    repr=repr,
    hash=hash,
    compare=compare,
    metadata=metadata,
  )


def treenode_field(
  *,
  default: tp.Any = dataclasses.MISSING,
  default_factory: tp.Any = dataclasses.MISSING,
  init: bool = True,
  repr: bool = True,
  hash: tp.Optional[bool] = None,
  compare: bool = True,
  metadata: tp.Optional[tp.Mapping[str, tp.Any]] = None,
):
  if metadata is None:
    metadata = {}
  else:
    metadata = dict(metadata)

  if 'nnx_variable_constructor' in metadata:
    raise ValueError("'nnx_variable_constructor' found in metadata")

  metadata['nnx_variable_constructor'] = lambda value: pytreelib.TreeNode(value)

  return field(
    default=default,
    default_factory=default_factory,
    init=init,
    repr=repr,
    hash=hash,
    compare=compare,
    metadata=metadata,
  )


def variable_field(
  variable_type: tp.Type[variables.Variable[tp.Any]],
  *,
  default: tp.Any = dataclasses.MISSING,
  default_factory: tp.Any = dataclasses.MISSING,
  init: bool = True,
  repr: bool = True,
  hash: tp.Optional[bool] = None,
  compare: bool = True,
  metadata: tp.Optional[tp.Mapping[str, tp.Any]] = None,
) -> tp.Any:
  if metadata is None:
    metadata = {}
  else:
    metadata = dict(metadata)

  if 'nnx_variable_constructor' in metadata:
    raise ValueError("'nnx_variable_constructor' found in metadata")

  metadata['nnx_variable_constructor'] = lambda value: variable_type(value)

  return field(
    default=default,
    default_factory=default_factory,
    init=init,
    repr=repr,
    hash=hash,
    compare=compare,
    metadata=metadata,
  )


def param_field(
  default: tp.Any = dataclasses.MISSING,
  *,
  default_factory: tp.Any = dataclasses.MISSING,
  init: bool = True,
  repr: bool = True,
  hash: tp.Optional[bool] = None,
  compare: bool = True,
  metadata: tp.Optional[tp.Mapping[str, tp.Any]] = None,
) -> tp.Any:
  return variable_field(
    variables.Param,
    default=default,
    default_factory=default_factory,
    init=init,
    repr=repr,
    hash=hash,
    compare=compare,
    metadata=metadata,
  )


@tp.overload
def dataclass(cls: tp.Type[A]) -> tp.Type[A]:
  ...


@tp.overload
def dataclass(
  *,
  init: bool = True,
  repr: bool = True,
  eq: bool = True,
  order: bool = False,
  unsafe_hash: bool = False,
  frozen: bool = False,
) -> tp.Callable[[tp.Type[A]], tp.Type[A]]:
  ...


@tpe.dataclass_transform(
  field_specifiers=(
    field,
    treenode_field,
    variable_field,
    param_field,
  )
)
def dataclass(
  cls: tp.Optional[tp.Type[A]] = None,
  init: bool = True,
  repr: bool = True,
  eq: bool = True,
  order: bool = False,
  unsafe_hash: bool = False,
  frozen: bool = False,
) -> tp.Union[tp.Type[A], tp.Callable[[tp.Type[A]], tp.Type[A]]]:
  def decorator(cls: tp.Type[A]) -> tp.Type[A]:
    cls = dataclasses.dataclass(
      init=init,
      repr=repr,
      eq=eq,
      order=order,
      unsafe_hash=unsafe_hash,
      frozen=frozen,
    )(cls)
    if issubclass(cls, nnx.Module):

      def hash_fn(module: nnx.Module):
        return hash(module._module__state.id)

      cls.__hash__ = hash_fn

    return cls

  if cls is None:
    return decorator

  return decorator(cls)



# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Definition of the GNN model."""

from typing import Callable, Sequence

from flax import linen as nn
import jax.numpy as jnp
import jraph


def add_graphs_tuples(
    graphs: jraph.GraphsTuple, other_graphs: jraph.GraphsTuple
) -> jraph.GraphsTuple:
  """Adds the nodes, edges and global features from other_graphs to graphs."""
  return graphs._replace(
      nodes=graphs.nodes + other_graphs.nodes,
      edges=graphs.edges + other_graphs.edges,
      globals=graphs.globals + other_graphs.globals,
  )


class MLP(nn.Module):
  """A multi-layer perceptron."""

  feature_sizes: Sequence[int]
  dropout_rate: float = 0
  deterministic: bool = True
  activation: Callable[[jnp.ndarray], jnp.ndarray] = nn.relu

  @nn.compact
  def __call__(self, inputs):
    x = inputs
    for size in self.feature_sizes:
      x = nn.Dense(features=size)(x)
      x = self.activation(x)
      x = nn.Dropout(rate=self.dropout_rate, deterministic=self.deterministic)(
          x
      )
    return x


class GraphNet(nn.Module):
  """A complete Graph Network model defined with Jraph."""

  latent_size: int
  num_mlp_layers: int
  message_passing_steps: int
  output_globals_size: int
  dropout_rate: float = 0
  skip_connections: bool = True
  use_edge_model: bool = True
  layer_norm: bool = True
  deterministic: bool = True

  @nn.compact
  def __call__(self, graphs: jraph.GraphsTuple) -> jraph.GraphsTuple:
    # We will first linearly project the original features as 'embeddings'.
    embedder = jraph.GraphMapFeatures(
        embed_node_fn=nn.Dense(self.latent_size),
        embed_edge_fn=nn.Dense(self.latent_size),
        embed_global_fn=nn.Dense(self.latent_size),
    )
    processed_graphs = embedder(graphs)

    # Now, we will apply a Graph Network once for each message-passing round.
    mlp_feature_sizes = [self.latent_size] * self.num_mlp_layers
    for _ in range(self.message_passing_steps):
      if self.use_edge_model:
        update_edge_fn = jraph.concatenated_args(
            MLP(
                mlp_feature_sizes,
                dropout_rate=self.dropout_rate,
                deterministic=self.deterministic,
            )
        )
      else:
        update_edge_fn = None

      update_node_fn = jraph.concatenated_args(
          MLP(
              mlp_feature_sizes,
              dropout_rate=self.dropout_rate,
              deterministic=self.deterministic,
          )
      )
      update_global_fn = jraph.concatenated_args(
          MLP(
              mlp_feature_sizes,
              dropout_rate=self.dropout_rate,
              deterministic=self.deterministic,
          )
      )

      graph_net = jraph.GraphNetwork(
          update_node_fn=update_node_fn,
          update_edge_fn=update_edge_fn,
          update_global_fn=update_global_fn,
      )

      if self.skip_connections:
        processed_graphs = add_graphs_tuples(
            graph_net(processed_graphs), processed_graphs
        )
      else:
        processed_graphs = graph_net(processed_graphs)

      if self.layer_norm:
        processed_graphs = processed_graphs._replace(
            nodes=nn.LayerNorm()(processed_graphs.nodes),
            edges=nn.LayerNorm()(processed_graphs.edges),
            globals=nn.LayerNorm()(processed_graphs.globals),
        )

    # Since our graph-level predictions will be at globals, we will
    # decode to get the required output logits.
    decoder = jraph.GraphMapFeatures(
        embed_global_fn=nn.Dense(self.output_globals_size)
    )
    processed_graphs = decoder(processed_graphs)

    return processed_graphs


class GraphConvNet(nn.Module):
  """A Graph Convolution Network + Pooling model defined with Jraph."""

  latent_size: int
  num_mlp_layers: int
  message_passing_steps: int
  output_globals_size: int
  dropout_rate: float = 0
  skip_connections: bool = True
  layer_norm: bool = True
  deterministic: bool = True
  pooling_fn: Callable[
      [jnp.ndarray, jnp.ndarray, jnp.ndarray],  # pytype: disable=annotation-type-mismatch  # jax-ndarray
      jnp.ndarray,
  ] = jraph.segment_mean

  def pool(self, graphs: jraph.GraphsTuple) -> jraph.GraphsTuple:
    """Pooling operation, taken from Jraph."""

    # Equivalent to jnp.sum(n_node), but JIT-able.
    sum_n_node = graphs.nodes.shape[0]  # pytype: disable=attribute-error  # jax-ndarray
    # To aggregate nodes from each graph to global features,
    # we first construct tensors that map the node to the corresponding graph.
    # Example: if you have `n_node=[1,2]`, we construct the tensor [0, 1, 1].
    n_graph = graphs.n_node.shape[0]
    node_graph_indices = jnp.repeat(
        jnp.arange(n_graph),
        graphs.n_node,
        axis=0,
        total_repeat_length=sum_n_node,
    )
    # We use the aggregation function to pool the nodes per graph.
    pooled = self.pooling_fn(graphs.nodes, node_graph_indices, n_graph)  # pytype: disable=wrong-arg-types  # jax-ndarray
    return graphs._replace(globals=pooled)

  @nn.compact
  def __call__(self, graphs: jraph.GraphsTuple) -> jraph.GraphsTuple:
    # We will first linearly project the original node features as 'embeddings'.
    embedder = jraph.GraphMapFeatures(embed_node_fn=nn.Dense(self.latent_size))
    processed_graphs = embedder(graphs)

    # Now, we will apply the GCN once for each message-passing round.
    for _ in range(self.message_passing_steps):
      mlp_feature_sizes = [self.latent_size] * self.num_mlp_layers
      update_node_fn = jraph.concatenated_args(
          MLP(
              mlp_feature_sizes,
              dropout_rate=self.dropout_rate,
              deterministic=self.deterministic,
          )
      )
      graph_conv = jraph.GraphConvolution(
          update_node_fn=update_node_fn, add_self_edges=True
      )

      if self.skip_connections:
        processed_graphs = add_graphs_tuples(
            graph_conv(processed_graphs), processed_graphs
        )
      else:
        processed_graphs = graph_conv(processed_graphs)

      if self.layer_norm:
        processed_graphs = processed_graphs._replace(
            nodes=nn.LayerNorm()(processed_graphs.nodes),
        )

    # We apply the pooling operation to get a 'global' embedding.
    processed_graphs = self.pool(processed_graphs)

    # Now, we decode this to get the required output logits.
    decoder = jraph.GraphMapFeatures(
        embed_global_fn=nn.Dense(self.output_globals_size)
    )
    processed_graphs = decoder(processed_graphs)

    return processed_graphs



# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for flax.traceback_util."""

import contextlib
import sys
import traceback

import jax
from absl.testing import absltest
from jax import numpy as jnp
from jax import random
from jax._src import traceback_util as jax_traceback_util

from flax import linen as nn
from flax import traceback_util

# pylint: disable=arguments-differ,protected-access, g-wrong-blank-lines

# __tracebackhide__ is a python >=3.7 feature.
TRACEBACKHIDE_SUPPORTED = tuple(sys.version_info)[:3] >= (3, 7, 0)

EXPECTED_FILES = (__file__, contextlib.__spec__.origin)


class TracebackTest(absltest.TestCase):
  def test_exclusion_list(self):
    traceback_util.show_flax_in_tracebacks()
    exclusion_len_wo_flax = len(jax_traceback_util._exclude_paths)
    traceback_util.hide_flax_in_tracebacks()
    exclusion_len_w_flax = len(jax_traceback_util._exclude_paths)
    self.assertLen(
      traceback_util._flax_exclusions,
      exclusion_len_w_flax - exclusion_len_wo_flax,
    )

  def test_simple_exclusion_tracebackhide(self):
    if not TRACEBACKHIDE_SUPPORTED:
      return

    class Test1(nn.Module):
      @nn.remat
      @nn.compact
      def __call__(self, x):
        return Test2()(x)

    class Test2(nn.Module):
      @nn.jit
      @nn.compact
      def __call__(self, x):
        raise ValueError('error here.')
        return x  # pylint: disable=unreachable

    traceback_util.hide_flax_in_tracebacks()
    jax.config.update('jax_traceback_filtering', 'tracebackhide')

    key = random.key(0)
    try:
      nn.jit(Test1)().init(key, jnp.ones((5, 3)))
    except ValueError as e:
      tb = e.__traceback__

    filtered_frames = 0
    unfiltered_frames = 0

    for f, _ in traceback.walk_tb(tb):
      if '__tracebackhide__' not in f.f_locals:
        self.assertIn(f.f_code.co_filename, EXPECTED_FILES)
        filtered_frames += 1
      unfiltered_frames += 1

    self.assertEqual(filtered_frames, 3)
    self.assertGreater(unfiltered_frames, filtered_frames)

  def test_simple_exclusion_remove_frames(self):
    class Test1(nn.Module):
      @nn.remat
      @nn.compact
      def __call__(self, x):
        return Test2()(x)

    class Test2(nn.Module):
      @nn.jit
      @nn.compact
      def __call__(self, x):
        raise ValueError('error here.')
        return x  # pylint: disable=unreachable

    traceback_util.hide_flax_in_tracebacks()
    jax.config.update('jax_traceback_filtering', 'remove_frames')

    key = random.key(0)
    try:
      nn.jit(Test1)().init(key, jnp.ones((5, 3)))
    except ValueError as e:
      tb_filtered = e.__traceback__
      tb_unfiltered = e.__cause__.__traceback__
      e_cause = e.__cause__

    self.assertIsInstance(e_cause, jax_traceback_util.UnfilteredStackTrace)

    filtered_frames = 0
    for _, _ in traceback.walk_tb(tb_filtered):
      filtered_frames += 1

    unfiltered_frames = 0
    for _, _ in traceback.walk_tb(tb_unfiltered):
      unfiltered_frames += 1

    self.assertEqual(filtered_frames, 3)
    self.assertGreater(unfiltered_frames, filtered_frames)

  def test_dynamic_exclusion(self):
    if not TRACEBACKHIDE_SUPPORTED:
      return

    class Test1(nn.Module):
      @nn.remat
      @nn.compact
      def __call__(self, x):
        return Test2()(x)

    class Test2(nn.Module):
      @nn.jit
      @nn.compact
      def __call__(self, x):
        raise ValueError('error here.')
        return x  # pylint: disable=unreachable

    key = random.key(0)

    traceback_util.show_flax_in_tracebacks()
    jax.config.update('jax_traceback_filtering', 'off')
    try:
      nn.jit(Test1)().init(key, jnp.ones((5, 3)))
    except ValueError as e:
      tb_all = e.__traceback__

    traceback_util.hide_flax_in_tracebacks()
    jax.config.update('jax_traceback_filtering', 'tracebackhide')
    try:
      nn.jit(Test1)().init(key, jnp.ones((5, 3)))
    except ValueError as e:
      tb_no_flax = e.__traceback__

    traceback_util.show_flax_in_tracebacks()
    jax.config.update('jax_traceback_filtering', 'tracebackhide')
    try:
      nn.jit(Test1)().init(key, jnp.ones((5, 3)))
    except ValueError as e:
      tb_w_flax = e.__traceback__

    filtered_frames_all = 0
    unfiltered_frames_all = 0
    for f, _ in traceback.walk_tb(tb_all):
      if '__tracebackhide__' not in f.f_locals:
        unfiltered_frames_all += 1
      else:
        filtered_frames_all += 1

    filtered_frames_no_flax = 0
    unfiltered_frames_no_flax = 0
    for f, _ in traceback.walk_tb(tb_no_flax):
      if '__tracebackhide__' not in f.f_locals:
        self.assertIn(f.f_code.co_filename, EXPECTED_FILES)
        unfiltered_frames_no_flax += 1
      else:
        filtered_frames_no_flax += 1

    filtered_frames_w_flax = 0
    unfiltered_frames_w_flax = 0
    for f, _ in traceback.walk_tb(tb_w_flax):
      if '__tracebackhide__' not in f.f_locals:
        unfiltered_frames_w_flax += 1
      else:
        filtered_frames_w_flax += 1

    self.assertEqual(
      unfiltered_frames_all + filtered_frames_all,
      unfiltered_frames_w_flax + filtered_frames_w_flax,
    )
    self.assertEqual(
      unfiltered_frames_all + filtered_frames_all,
      unfiltered_frames_no_flax + filtered_frames_no_flax,
    )
    self.assertEqual(unfiltered_frames_no_flax, 3)
    self.assertGreater(unfiltered_frames_all, unfiltered_frames_w_flax)
    self.assertGreater(unfiltered_frames_w_flax, unfiltered_frames_no_flax)


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Benchmark for the ImageNet example."""

import time

from absl import flags
from absl.testing import absltest
from absl.testing.flagsaver import flagsaver
from flax.testing import Benchmark
import jax
import numpy as np

# Local imports.
import main
from configs import v100_x8_mixed_precision as config_lib


# Parse absl flags test_srcdir and test_tmpdir.
jax.config.parse_flags_with_absl()

FLAGS = flags.FLAGS


class ImagenetBenchmark(Benchmark):
  """Benchmarks for the ImageNet Flax example."""

  @flagsaver
  def _test_8x_v100_half_precision(
      self, num_epochs: int, min_accuracy, max_accuracy
  ):
    """Utility to benchmark ImageNet on 8xV100 GPUs. Use in your test func."""
    # Prepare and set flags defined in main.py.
    config = config_lib.get_config()
    config.num_epochs = num_epochs
    workdir = self.get_tmp_model_dir()

    FLAGS.workdir = workdir
    FLAGS.config = config

    start_time = time.time()
    main.main([])
    benchmark_time = time.time() - start_time
    summaries = self.read_summaries(workdir)

    # Summaries contain all the information necessary for the regression
    # metrics.
    wall_time, _, eval_accuracy = zip(*summaries['eval_accuracy'])
    wall_time = np.array(wall_time)
    sec_per_epoch = np.mean(wall_time[1:] - wall_time[:-1])
    end_accuracy = eval_accuracy[-1]

    # Assertions are deferred until the test finishes, so the metrics are
    # always reported and benchmark success is determined based on *all*
    # assertions.
    self.assertBetween(end_accuracy, min_accuracy, max_accuracy)

    # Use the reporting API to report single or multiple metrics/extras.
    self.report_wall_time(benchmark_time)
    self.report_metrics(
        {'sec_per_epoch': sec_per_epoch, 'accuracy': end_accuracy}
    )

  def test_8x_v100_half_precision_short(self):
    """Run ImageNet on 8x V100 GPUs in half precision for 2 epochs."""
    self._test_8x_v100_half_precision(
        num_epochs=2, min_accuracy=0.06, max_accuracy=0.09
    )
    self.report_extras({
        'description': 'Short (2 epochs) 8 x V100 test for ImageNet ResNet50.',
        'model_name': 'resnet50',
        'parameters': 'hp=true,bs=2048,num_epochs=2',
        'implementation': 'linen',
    })

  def test_8x_v100_half_precision_full(self):
    """Run ImageNet on 8x V100 GPUs in half precision for full 90 epochs."""
    self._test_8x_v100_half_precision(
        num_epochs=90, min_accuracy=0.76, max_accuracy=0.77
    )
    self.report_extras({
        'description': 'Full (90 epochs) 8 x V100 test for ImageNet ResNet50.',
        'model_name': 'resnet50',
        'parameters': 'hp=true,bs=2048,num_epochs=90',
        'implementation': 'linen',
    })


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for linen_meta."""

import jax
from absl.testing import absltest
from jax import numpy as jnp
from jax import random
from jax.experimental import mesh_utils
from jax.sharding import Mesh, PartitionSpec

from flax import linen as nn


class LinenMetaTest(absltest.TestCase):
  def test_boxed_param(self):
    class Bar(nn.Module):
      @nn.compact
      def __call__(mdl_self, x):  # pylint: disable=no-self-argument
        kernel_init = nn.with_partitioning(
          nn.initializers.ones_init(), ('in', 'out')
        )
        kernel = mdl_self.param('kernel', kernel_init, (x.shape[-1], 2))
        kernel_box = mdl_self.get_variable('params', 'kernel')
        self.assertIsInstance(kernel_box, nn.Partitioned)
        self.assertEqual(kernel_box.names, ('in', 'out'))
        return x @ kernel

    class Foo(nn.Module):
      @nn.compact
      def __call__(self, xs):
        return nn.vmap(
          Bar,
          in_axes=0,
          variable_axes={'params': 0},
          split_rngs={'params': True},
          metadata_params={nn.PARTITION_NAME: 'batch'},
        )(name='bar')(xs)

    m = Foo()
    variables = m.init(random.key(0), jnp.zeros((8, 3)))
    self.assertEqual(
      variables['params']['bar']['kernel'].names, ('batch', 'in', 'out')
    )

  def test_boxed_variable(self):
    class Bar(nn.Module):
      @nn.compact
      def __call__(mdl_self, x):  # pylint: disable=no-self-argument
        kernel_init = nn.with_partitioning(
          nn.initializers.ones_init(), ('in', 'out')
        )
        kernel = mdl_self.variable(
          'params',
          'kernel',
          kernel_init,
          mdl_self.make_rng('params'),
          (x.shape[-1], 2),
        )
        kernel.value += 1.0
        self.assertEqual(kernel.value.sum(), kernel.value.size * 2)
        kernel_box = mdl_self.get_variable('params', 'kernel')
        self.assertIsInstance(kernel_box, nn.Partitioned)
        self.assertEqual(kernel_box.names, ('in', 'out'))
        return x @ kernel.value

    class Foo(nn.Module):
      @nn.compact
      def __call__(self, xs):
        return nn.vmap(
          Bar,
          in_axes=0,
          variable_axes={'params': 0},
          split_rngs={'params': True},
          metadata_params={nn.PARTITION_NAME: 'batch'},
        )(name='bar')(xs)

    m = Foo()
    variables = m.init(random.key(0), jnp.zeros((8, 3)))
    self.assertEqual(
      variables['params']['bar']['kernel'].names, ('batch', 'in', 'out')
    )

  # def test_boxed_variable(self):
  #   def f(scope, xs):
  #     def g(scope, x):
  # kernel_init = nn.with_partitioning(nn.initializers.ones_init(),
  #                                      ('in', 'out'))
  # kernel = scope.variable('params', 'kernel', kernel_init,
  #                         scope.make_rng('params'), (x.shape[-1], 2))
  # kernel.value += 1.
  # self.assertEqual(kernel.value.sum(), kernel.value.size * 2)
  # kernel_box = scope.get_variable('params', 'kernel')
  # self.assertIsInstance(kernel_box, nn.Partitioned)
  # self.assertEqual(kernel_box.names, ('in', 'out'))
  # return x @ kernel.value

  #     nn.vmap(
  #         g, in_axes=0,
  #         variable_axes={'params': 0}, split_rngs={'params': True},
  #         metadata_params={nn.PARTITION_NAME: 'batch'})(scope, xs)

  #   _, variables = init(f)(random.key(0), jnp.zeros((8, 3)))
  #   self.assertEqual(variables['params']['kernel'].names,
  #                    ('batch', 'in', 'out'))

  def test_pjit_scan_over_layers(self):
    class MLP(nn.Module):
      hidden_size: int

      @nn.compact
      def __call__(self, x):
        ki = nn.linear.default_kernel_init
        h = nn.Dense(
          self.hidden_size,
          kernel_init=nn.with_partitioning(ki, ('data', 'model')),
        )(x)
        h = nn.relu(h)
        return nn.Dense(
          x.shape[-1], kernel_init=nn.with_partitioning(ki, ('model', 'data'))
        )(h)

    class Model(nn.Module):
      @nn.compact
      def __call__(self, x):
        def body(_, c):
          c = MLP(512)(c)
          return c, ()

        c, _ = nn.scan(
          body,
          variable_axes={'params': 0},
          split_rngs={'params': 0},
          length=8,
          metadata_params={nn.PARTITION_NAME: None},
        )(self, x)
        return c

    devs = mesh_utils.create_device_mesh((jax.device_count(), 1))
    mesh = Mesh(devs, ['data', 'model'])
    model = Model()
    x = jnp.ones((8, 128))
    spec = nn.get_partition_spec(jax.eval_shape(model.init, random.key(0), x))
    self.assertEqual(
      spec,
      {
        'params': {
          'MLP_0': {
            'Dense_0': {
              'bias': PartitionSpec(),
              'kernel': PartitionSpec(None, 'data', 'model'),
            },
            'Dense_1': {
              'bias': PartitionSpec(),
              'kernel': PartitionSpec(None, 'model', 'data'),
            },
          },
        },
      },
    )
    x_spec = PartitionSpec('data', 'model')
    f = lambda x: jax.sharding.NamedSharding(mesh, x)
    key_spec = PartitionSpec()
    init_fn = jax.jit(
      model.init,
      in_shardings=jax.tree_map(f, (key_spec, x_spec)),
      out_shardings=jax.tree_map(f, spec),
    )
    variables = init_fn(random.key(0), x)
    apply_fn = jax.jit(
      model.apply,
      in_shardings=jax.tree_map(f, (spec, x_spec)),
      out_shardings=jax.tree_map(f, x_spec),
    )
    y = apply_fn(variables, x)
    self.assertEqual(y.shape, (8, 128))


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Default Hyperparameter configuration."""

import ml_collections


def get_config():
  """Get the default hyperparameter configuration."""
  config = ml_collections.ConfigDict()

  # Path to load or store sentencepiece vocab file.
  config.vocab_path = None

  # Vocabulary size if `vocab_path` is not given.
  config.vocab_size = 30_000

  config.max_corpus_chars = 10**7

  # Name of TFDS translation dataset to use.
  config.dataset_name = 'lm1b'

  # Optional name of TFDS translation dataset to use for evaluation.
  config.eval_dataset_name = 'lm1b'
  config.eval_split = 'test'

  # Per device batch size for training.
  config.per_device_batch_size = 32

  # Per device batch size for training.
  config.eval_per_device_batch_size = 32

  # Sampling temperature for language model inference.
  config.sampling_temperature = 0.6

  # Top k cutoff for logit sampling. If 0 then no top-k cutoff is used.
  config.sampling_top_k = 20

  config.num_train_steps = 500_000

  # Number of steps to take during evaluation. Large enough to evaluate all.
  # Large enough to evaluate all samples: 306_688 / (32 * 8) = 1198
  config.num_eval_steps = 2_000
  # Number of steps to generate predictions.
  # -1 will use the whole eval dataset.
  config.num_predict_steps = -1

  # Base learning rate.
  config.learning_rate = 0.0016

  # Linear learning rate warmup.
  config.warmup_steps = 1000

  # Cross entropy loss label smoothing.
  config.label_smoothing = 0.0

  # Decay factor for AdamW style weight decay.
  config.weight_decay = 0.1

  # Maximum length cutoff for training examples.
  config.max_target_length = 128
  # Maximum length cutoff for eval examples.
  config.max_eval_target_length = 512
  # Maximum length cutoff for predicted tokens.
  config.max_predict_length = 50

  # Final logit transform uses embedding matrix transpose.
  config.logits_via_embedding = False

  # Number of transformer layers.
  config.num_layers = 6

  # Size of query/key/value for attention.
  config.qkv_dim = 512
  # Size of embeddings.
  config.emb_dim = 512
  # Size of the MLP.
  config.mlp_dim = 2048

  # Number of attention heads.
  config.num_heads = 8

  # Dropout rate.
  config.dropout_rate = 0.1

  # Attention dropout rate.
  config.attention_dropout_rate = 0.1

  # Whether to save model checkpoints.
  config.save_checkpoints = True
  # Whether to restore from existing model checkpoints.
  config.restore_checkpoints = True

  # Save a checkpoint every these number of steps.
  config.checkpoint_every_steps = 10_000
  # Frequency of eval during training, e.g. every 1_000 steps.
  config.eval_every_steps = 1_000

  # Use bfloat16 mixed precision training instead of float32.
  config.use_bfloat16 = True

  # Integer for PRNG random seed.
  config.seed = 0

  # Prompt for language model sampling,
  # taken from MaxText (https://github.com/google/maxtext/blob/main/MaxText/configs/base.yml).
  config.prompts = 'I love to '

  # Parallelism
  config.mesh_axes = ['data', 'fsdp', 'tensor']
  config.logical_axis_rules = [
      ['activation_batch', ['data', 'fsdp']],
      ['activation_length', ['data', 'fsdp']],
      ['activation_embed', 'tensor'],
      ['activation_mlp', 'tensor'],
      ['activation_heads', 'tensor'],
      ['activation_kv', 'tensor'],
      ['activation_vocab', 'tensor'],
      ['mlp', 'tensor'],
      ['vocab', 'tensor'],
      ['embed', 'fsdp'],
      ['heads', 'tensor'],
  ]
  config.data_sharding = ['data']

  # One axis for each parallelism type may hold a placeholder (-1)
  # value to auto-shard based on available slices and devices.
  # By default, product of the DCN axes should equal number of slices
  # and product of the ICI axes should equal number of devices per slice.
  # ICI (Inter-Chip Interconnection): A high-speed connection between
  # sets of TPU chips, which form the TPU network.
  # DCN (Data Center Network): A connection between the TPU networks;
  # not as fast as ICI.
  # ICI has around 100x the bandwidth of DCN, but it is not a general
  # purpose connection, which is why DCN is necessary for scaling to
  # extremely large ML models.
  config.dcn_data_parallelism = -1  # recommended DCN axis to be auto-sharded
  config.dcn_fsdp_parallelism = 1
  config.dcn_tensor_parallelism = 1
  config.ici_data_parallelism = 1
  config.ici_fsdp_parallelism = -1  # recommended ICI axis to be auto-sharded
  config.ici_tensor_parallelism = 1

  return config
